--Change log:
--  2010-02-10: created ai8
--
--  2010-01-24: created ai7
--
--  2010-01-16: created ai6; trying iterative deepening
--
--  2010-01-12: created ai5; use priority queue to analyze positions,
--    consider ability to contribute blocks important
--
--  2010-01-10: created ai4; use "planner2" personality when calculating
--    the end result of a position; use breadth-first searching of game
--    tree

local TUNABLE = {
		WANT_MONEY_THRESHOLD = 6,   --[0,inf]
		VALUE_OF_CARPENTER = 2,     --[0,inf]
		VALUE_OF_WOODEN_PRODUCER = 2,
		VALUE_OF_MASON = 2,
		VALUE_OF_OTHER_WOODEN = 1,
		VALUE_OF_STONE_PRODUCER = 3,
		VALUE_OF_OTHER_STONE = 1,
	}
local ai = {}
local fallback_ai = load_ai_personality("planner4.lua")
local quick_ai = {}

local AI_SEARCH_BREADTH = 3
local AI_SEARCH_DEPTH = 1
ai.VERBOSE = false

local function estimate_maximum_contribution(r)
	return math.max(
			math.min(r.food, r.wood, r.stone + r.cloth),
			math.min(r.food, r.stone, r.wood + r.cloth),
			math.min(r.food, r.cloth, r.wood + r.stone)
			)
end

local function estimate_num_rounds_left()
	if state.castle_phase > 3 then
		return 0
	end

	local bailiff_left = CASTLE_BAILIFF_THRESHOLDS[3] - state.bailiff 
	local castle_left = CASTLE_CONTRIBUTION_THRESHOLDS[3] - state.contribution_count
	if castle_left <= 4 then return 1 end
	if castle_left <= 7 then return 2 end

	local rounds_left = bailiff_left * 3 / 4
	if state.castle_phase == 3 then
		rounds_left = math.max(rounds_left, 5)
	elseif state.castle_phase == 2 then
		rounds_left = math.max(rounds_left, 10)
	elseif state.castle_phase == 1 then
		rounds_left = math.max(rounds_left, 15)
	end

	return rounds_left
end

local EVALUATION
function ai.evaluate_position_detailed()
	ai.evaluate_position(1)
	for pid,player in pairs(state.players) do
		local tmp = string.format("%-9s: ", player.name)
		local sum = 0
		for c,x in pairs(EVALUATION[pid]) do
			tmp = tmp .. string.format(" %s=%.1f", c, x)
			sum = sum + x
		end
		tmp = tmp .. string.format(" sum=%.1f", sum)
		print(tmp)
	end
end

local function estimate_value_of_building(position, bldgref)
	--estimate how many rounds are left
	local rounds_left = estimate_num_rounds_left()

	if bldgref.btype == "residence" then
		return math.min(3,rounds_left) * 3
	end

	--estimate how many rounds before the building can be used
	local rounds_wait = math.max(0, position - state.bailiff)
	local factor = (rounds_left - rounds_wait) / 12

	if bldgref == WOOD_BUILDINGS[1]
		or bldgref == WOOD_BUILDINGS[2]
		or bldgref == WOOD_BUILDINGS[3]
		or bldgref == WOOD_BUILDINGS[4]
		then
		return factor * TUNABLE.VALUE_OF_WOODEN_PRODUCER
	elseif bldgref == MASON then
		return factor * TUNABLE.VALUE_OF_MASON
	elseif bldgref == STONE_BUILDINGS[1]
		or bldgref == STONE_BUILDINGS[2]
		or bldgref == STONE_BUILDINGS[3]
		then
		return factor * TUNABLE.VALUE_OF_STONE_PRODUCER
	elseif bldgref.btype == "wood" then
		return factor * TUNABLE.VALUE_OF_OTHER_WOODEN
	elseif bldgref.btype == "stone" then
		return factor * TUNABLE.VALUE_OF_OTHER_STONE
	else
		return 0
	end
end

local function value_of_favor(player_id)
	return 3
end

local function player_is_at_building(player_id, building)
	for i,pid in ipairs(building.occupants) do
		if pid == player_id then
			return true
		end
	end
	return false
end

local function evaluate_position_at_round_end(whoami, initial)
	local rounds_left = estimate_num_rounds_left()
	local est_score = {}
	local sum_all_players = 0

	EVALUATION = {}
	for player_id,me in pairs(state.players) do

		local eval = initial[player_id] or {}
		EVALUATION[player_id] = eval

		eval.vp = me.score

		if rounds_left > 0 then
			local r = {
				food = me.resources.food,
				wood = me.resources.wood,
				stone = me.resources.stone,
				cloth = me.resources.cloth,
				gold = me.resources.gold,
				money = me.resources.money,
				points = me.score,
				}

			-- value of unused gold
			eval.gold = 4 * r.gold

			-- value of potential castle blocks
			local contrib = estimate_maximum_contribution(r)
			local worth_of_block = 4 + (rounds_left / 5)
			eval.contrib = worth_of_block * contrib
			
			-- value of other resources
			eval.resources = math.max(0,
				(r.food
				+ r.wood
				+ r.stone
				+ r.cloth) - 3 * contrib)

			-- TODO- value of having certain number of contributions
			-- in current castle phase

			-- value of money in the bank
			--  note: money not worth as much when you have a bunch
			--  note: if less than 2 rounds left, then money not
			--        worth as much either
			eval.money = math.sqrt(math.min(3,rounds_left) * 2 * r.money)

			-- value of going early in the turn order
			if rounds_left > 0 and #state.order > 2 then
				eval.order = 1.5 - .5 * position_in_turn_order(whoami)
			end

		--	-- value of occupying the inn
		--	if rounds_left > 0 then
		--		eval.inn = 0
		--	end

			-- value of owning buildings that other players use
			eval.buildings = 0
			for i = BRIDGE_POSITION+1, LAST_EMPTY_POSITION do
				if state.buildings[i].owner == player_id then
					local bldgref = state.buildings[i].ref
					local bldg_value = estimate_value_of_building(i,bldgref)
					eval.buildings = eval.buildings + bldg_value
				end
			end
		end --if any rounds left

		local sum = 0
		for i,x in pairs(eval) do
			sum = sum + x
		end

		est_score[player_id] = sum
		sum_all_players = sum_all_players + sum

	end --loop through players

	local avg_opponent_score = (sum_all_players - est_score[whoami])
			/ (#state.order - 1)

	return est_score[whoami] - avg_opponent_score
end

function ai.evaluate_position(whoami)

	local old_state = state
	state = clone_game_state(old_state)

	local orig_no_pass_reward = NO_PASS_REWARD
	NO_PASS_REWARD = true
	local adjustment = {}
	if state.phase == PHASE_WORKER_PLACEMENT then
		for i,player in ipairs(state.players) do
			adjustment[i] = {
				unused_workers = -player.workers_left,
				}
		end
	end

	local principal_variation = play_random_game_to_next_round(quick_ai)
	local rv = evaluate_position_at_round_end(whoami, adjustment)

	NO_PASS_REWARD = orig_no_pass_reward
	state = old_state
	return rv, principal_variation
end

local whoami = 0

local principal_variation

function quick_ai.select_provost_move()
	local whoami = current_turn()
	local is_free = (state.active_building == MERCHANT_GUILD_POSITION)
	local max_movement = is_free and 3 or math.min(3,state.players[whoami].resources.money)
	if max_movement == 0 then
		return "noaction"
	end
	local last_my_worker = 0
	local last_opponent_worker = 0
	local need_to_save_coins = false
	for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do
		local bldg = state.buildings[i]
		if bldg.occupants[1] == whoami then
			last_my_worker = i
			if bldg.ref.id == "lawyer"
				or bldg.ref.id == "bank"
				or bldg.ref.id == "church"
				or bldg.ref.id == "wood-peddler"
				or bldg.ref == "fixed-peddler" then
				need_to_save_coins = true
			end
		elseif bldg.occupants[1] then
			last_opponent_worker = i
		end
	end
	if need_to_save_coins and not is_free then
		max_movement = math.min(3,state.players[whoami].resources.money - 1)
		if max_movement <= 0 then
			return "noaction"
		end
	end

		--TODO
end

local preferred_moves = {}
function quick_ai.select_move()
	local ph = current_phase()
	if ph == PHASE_WORKER_PLACEMENT then
		return "pass"
	--elseif ph == PHASE_MOVE_PROVOST then
	--	return quick_ai.select_provost_move()
	else
		local whoami = current_turn()
		if state.active_building > 0 and state.active_building ~= BRIDGE_POSITION then
			local key = string.format("%d,%d", whoami, state.active_building)
			local m = preferred_moves[key]
			if m and check_move(m) then
				return m
			end
			m = fallback_ai.select_move()
			preferred_moves[key] = m
			return m
		else
			return fallback_ai.select_move()
		end
	end
end

local function add_sorted(sorted_array, entry)

	for i = 1, #sorted_array do
		if sorted_array[i].score > entry.score then
			table.insert(sorted_array, i, entry)
			return
		end
	end
	table.insert(sorted_array, entry)

end

local function dump_choice(choice)
	local moves_string = choice.move
	local ee = choice.previous
	while ee.move do
		moves_string = ee.move .. "|" .. moves_string
		ee = ee.previous
	end
	if choice.principal_variation then
		moves_string = moves_string .. "   ("
		for i,m in ipairs(choice.principal_variation) do
			if i > 1 then moves_string = moves_string .. "|" end
			moves_string = moves_string .. m
		end
		moves_string = moves_string .. ")"
	end
	if string.len(moves_string) > 74 then
		moves_string = string.sub(moves_string, 1, 72) .. ".."
	end
	print(string.format("%5.1f %s", choice.score, moves_string))
end

local function debug_choices_heap(heap)
	for i = 1, math.min(#heap, 20) do
		local e = heap[#heap + 1 - i]
		dump_choice(e)
	end
end

local function process_choice(choice, replies)

	state = choice.state
	if caylus.is_end_of_game(state) then
		return choice.score, choice.principal_variation
	end

	local choices = all_available_moves()
	if #choices == 1 and #choice.principal_variation >= 1 then
		assert(choices[1] == choice.principal_variation[1])
		do_move(choices[1])

		local new_pv = {}
		for i=2,#choice.principal_variation do
			new_pv[i-1] = choice.principal_variation[i]
		end

		choice.state = nil
		local new_choice = {
			move = choices[1],
			principal_variation = new_pv,
			score = choice.score,
			state = state,
			strong_replies = {},
			previous = choice,
			}
		return process_choice(new_choice, {})
	end

		local _is_my_turn = is_my_turn(whoami) and 1 or -1
		local phase = current_phase()
		if phase == PHASE_MOVE_PROVOST then
			--forget replies for other moves
			replies = {}
		end

		local results = {}

		local function evaluate_next_move(choice, m)
			state = clone_game_state(choice.state)
			do_move(m)
			local this_score = ai.evaluate_position(whoami)
		--print(m)
		--ai.evaluate_position_detailed()
		--print("--")

			local e = {
				move = m,
				score = (this_score - choice.score) * _is_my_turn,
				previous = choice,
				strong_replies = {},
				state = state,
				}
			add_sorted(results, e)
			if not (replies[m] and replies[m] >= e.score) then
				replies[m] = e.score
			end
		end

		table.sort(choices,
			function(a,b) return (replies[b] or 0) < (replies[a] or 0) end
			)
		--shuffle(choices)

		local done_it = {}
		local done_count = 0

		--consider "new" replies only
		for j,m in ipairs(choices) do
			if not replies[m] then
				evaluate_next_move(choice, m)
				done_it[m] = 1
				done_count = done_count + 1
			end
		end

		local max_to_look_at = 6
		for j,m in ipairs(choices) do
			if done_count >= max_to_look_at then break end
			if not done_it[m] then
				evaluate_next_move(choice, m)
				done_it[m] = 1
				done_count = done_count + 1
			end
		end

		if ai.VERBOSE then
			print("next player's choices:")
			debug_choices_heap(results)
			print("--")
		end

		--figure out the move that offers the most resistance
		local strongest_reply = table.remove(results)
		choice.score = choice.score + strongest_reply.score * _is_my_turn

		--and record it as a "strong reply" in the parent node
		while strongest_reply and choice.previous.strong_replies[strongest_reply.move] do
			strongest_reply = table.remove(results)
		end
		if strongest_reply then
			choice.previous.strong_replies[strongest_reply.move] = strongest_reply
		end

		--dump_choice(e)

	return choice.score, choice.principal_variation

end

function ai.select_move()

	whoami = current_turn()
	preferred_moves = {}
	local saved_state = state

	local init_choice = {
		state = saved_state,
		strong_replies = {},
		}
	local choices = all_available_moves()
	if #choices == 1 then
		return choices[1]
	end

	local sorted_results = {}
	for i,m in ipairs(choices) do
		
		state = clone_game_state(saved_state)
		do_move(m)
		local this_score, pv = ai.evaluate_position(whoami)

		add_sorted(sorted_results,
			{
			move = m,
			principal_variation = pv,
			score = this_score,
			state = state,
			strong_replies = {},
			previous = init_choice,
			} )
	end

	if ai.VERBOSE then
		print("my choices:")
		debug_choices_heap(sorted_results)
		print("--")
	end

	local new_results = {}
	local replies = {}
	for i = 1,#sorted_results do

		if i > 8 then break end

		local choice = table.remove(sorted_results)
		if not choice then break end

		process_choice(choice, replies)
		add_sorted(new_results, choice)
	end

	if ai.VERBOSE then
		print("my choices, adjusted for opponent's reply:")
		debug_choices_heap(new_results)
		print("--")
	end

	state = saved_state
	local best_play = table.remove(new_results)
	return best_play.move
end

return ai
