local ai = load_ai_personality("commonguy.inc")

function ai.select_castle_move()
	local all_moves = all_available_moves()
	if #all_moves == 1 then
		return all_moves[1]
	end

	local options_by_count = {}
	for i,m in ipairs(all_moves) do
		local c = string.match(m, "^castle (.*)$")
		assert(c)
		if c == "nobuild" then
			--do nothing
		else
			local n = math.floor(string.len(c) / 3)
			assert(n >= 1)
			if not options_by_count[n] then
				options_by_count[n] = {}
			end
			table.insert(options_by_count[n], m)
		end
	end

	local my_best = #options_by_count
	assert(my_best >= 1)

	if state.castle_phase == 3 then
		--contribute as much as possible
		return options_by_count[my_best][1]
	end

	local best_from_followers = 0
	for i = state.turn + 1, #state.buildings[CASTLE_POSITION].occupants do
		local pid = state.buildings[CASTLE_POSITION].occupants[i]
		local m = ai.estimate_maximum_contribution(state.players[pid].resources)
		if m > best_from_followers then
			best_from_followers = m
		end
	end

	if state.turn > 1 and my_best <= state.best_contribution_amount then
		--one of my opponents already gave as many blocks as I can
		--contribute this turn.
		--So... contribute just one block, saving the rest for
		--a future turn.
		return options_by_count[1][1]
	end

	if best_from_followers >= 1 and my_best >= best_from_followers then
		--give just enough so that none of my opponents can beat
		--my contribution
		return options_by_count[best_from_followers][1]
	end

	if state.best_contribution_amount + 1 <= my_best then
		return options_by_count[state.best_contribution_amount + 1][1]
	end

	--give as many as possible
	return options_by_count[my_best][1]
end

function ai.value_of_resources_delta(ai, bag, orig_resources)
	if not orig_resources then
		local whoami = current_turn()
		local me = state.players[whoami]
		orig_resources = me.resources
	end

	return
		3.5 * bag.food * 4 / (4 + orig_resources.food) +
		2.5 * bag.wood * 4 / (4 + orig_resources.wood) +
		2.5 * bag.stone * 4 / (4 + orig_resources.stone) +
		2.5 * bag.cloth * 4 / (4 + orig_resources.cloth) +
		9 * bag.gold * 4 / (4 + orig_resources.gold)
end

function ai.select_worker_action()
	local bldgref = state.buildings[state.active_building].ref

	if state.turn == TURN_OWNER_ACTION then
		local s,m = ai:pick_building_owner_action(bldgref)
		return m
	end

	local s,m = ai:pick_building_pickup_action(bldgref)
	if s > 0 then
		return m
	end

	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

local function logisticfn(t)
	return 1 / (1 + math.exp(-t))
end

local function value_of_taking_castle(r, existing_occupants)
	local n = ai.estimate_maximum_contribution(r)
	if n >= 3 then
		return 3000
	elseif n >= 1 and #existing_occupants == 0 then
		return 3000
	end
	return 0
end

function ai.value_of_worker_at(i, r)
	local bldgref = assert(state.buildings[i].ref)

	local my_score = ai:pick_building_pickup_action(bldgref)
	if i == 2 and r.money < 4 then
		my_score = 5000
	end
	if i == 36 then
		my_score = value_of_taking_castle(r, state.buildings[i].occupants)
	end

	if i < CASTLE_POSITION then
		my_score = my_score * logisticfn(state.provost - i)
	end
	return my_score
end

function ai.select_worker_placement()
	local whoami = current_turn()
	local valid_moves = {}
	for i,m in ipairs(all_available_moves()) do
		valid_moves[m] = true
	end

	local me = state.players[whoami]
	local r = me.resources

	local best_move = "pass"
	local best_score = 0
	for i = 1, CASTLE_POSITION do
		local movestr = string.format("claim %d", i)
		if valid_moves[movestr] then
			local my_score = ai.value_of_worker_at(i, r)
			if my_score > best_score and my_score >= get_cost_to_place_worker(whoami, state.buildings[i]) then
				best_move = movestr
				best_score = my_score
			end
		end
	end
	return best_move
end

function ai.select_favor_move()
	local whoami = current_turn()
	local me = state.players[whoami]

	local favor_options = {
		vp = {},
		money = {},
		resources = {},
		builder = {},
		}
	for i,m in ipairs(all_available_moves()) do
		if string.match(m, "^favor (%d+)vp$") then
			table.insert(favor_options.vp, m)
		elseif string.match(m, "^favor (%d+)%$$") then
			table.insert(favor_options.money, m)
		end
	end

	if me.resources.money < 3 and favor_options.money[1] then
		return favor_options.money[1]
	elseif favor_options.vp[1] then
		return favor_options.vp[1]
	else
		return (all_available_moves())[1]
	end
end

return ai
