--Change log:
--  2010-02-11: created tree1
--
--  2010-02-10: created ai8
--
--  2010-01-24: created ai7
--
--  2010-01-16: created ai6; trying iterative deepening
--
--  2010-01-12: created ai5; use priority queue to analyze positions,
--    consider ability to contribute blocks important
--
--  2010-01-10: created ai4; use "planner2" personality when calculating
--    the end result of a position; use breadth-first searching of game
--    tree

local TUNABLE = {
		WANT_MONEY_THRESHOLD = 6,   --[0,inf]
	--these are estimates of how likely in any given round, if the building
	--is available, that someone will place a worker on it
		VALUE_OF_WOODEN_PRODUCER = 0.90,
		VALUE_OF_MASON =           0.45,
		VALUE_OF_LAWYER =          0.45,
		VALUE_OF_OTHER_WOODEN =    0.15,
		VALUE_OF_STONE_PRODUCER =  0.90,
		VALUE_OF_OTHER_STONE =     0.25,
		VALUE_OF_RESIDENCE =       4.00,
	--other tunables
		VALUE_OF_EARLY_TURN_ORDER = 1.0,
		VALUE_OF_RESOURCES_PARAM1 = 8,
		VALUE_OF_RESOURCES_PARAM2 = 7,
		VALUE_OF_MONEY_PARAM1 = 8,
		VALUE_OF_MONEY_PARAM2 = 5,
	}
local ai = {}
local fallback_ai = load_ai_personality("planner4.lua")
local quick_ai = {}

local AI_SEARCH_BREADTH = 3
local AI_SEARCH_DEPTH = 1
ai.VERBOSE = false
ai.TUNABLE = TUNABLE

--this variable gets reset between calls to select_move()
local context = {}

local function debug(message)
	if ai.VERBOSE then
		print(message)
	end
end

local function estimate_maximum_contribution(r)

	--FIXME: with 3f,2w,2s,2c then it is possible to make 3 contributions,
	--but this function does not recognize this

	return math.max(
			math.min(r.food, r.wood, r.stone + r.cloth),
			math.min(r.food, r.stone, r.wood + r.cloth),
			math.min(r.food, r.cloth, r.wood + r.stone)
			)
end

local function estimate_num_rounds_left()
	if state.castle_phase > 3 then
		return 0
	end
	local TYPICAL_NUMBER_OF_ROUNDS = 14.7

	local bailiff_left = CASTLE_BAILIFF_THRESHOLDS[3] - state.bailiff 
	local castle_left = CASTLE_CONTRIBUTION_THRESHOLDS[3] - state.contribution_count

	local bailiff_portion = (state.bailiff - BAILIFF_START_POSITION)
		/ (CASTLE_BAILIFF_THRESHOLDS[3] - BAILIFF_START_POSITION)
	local castle_portion = state.contribution_count / CASTLE_CONTRIBUTION_THRESHOLDS[3]

	local est_castle_rounds = castle_portion > 0 and state.round / castle_portion or 15
	local est_bailiff_rounds = bailiff_portion > 0 and state.round / bailiff_portion or 15
	local est_rounds = math.min(est_castle_rounds, est_bailiff_rounds)
	local est_rounds_left = est_rounds - state.round

	if bailiff_left <= 1 or castle_left <= 4 then
		return math.min(1, est_rounds_left)
	end
	if bailiff_left <= 3 or castle_left <= 7 then
		return math.min(2, est_rounds_left)
	end
	if state.castle_phase == 3 then
		return math.min(5, est_rounds_left)
	elseif state.castle_phase == 2 then
		local est1 = TYPICAL_NUMBER_OF_ROUNDS - (state.round - 1)
		local est2 = math.min(9, est_rounds_left)
		return (est1 + est2) / 2
	elseif state.castle_phase == 1 then
		return TYPICAL_NUMBER_OF_ROUNDS - (state.round - 1)
	end

	return est_rounds_left
end

function ai.estimate_rounds_left()
	return estimate_num_rounds_left()
end

local EVALUATION
local function dump_evaluation_details()
	for pid,player in pairs(state.players) do
		local tmp = ""
		local sum = 0
		for c,x in pairs(EVALUATION[pid]) do
			tmp = tmp .. string.format(" %s=%.1f", c, x)
			sum = sum + x
		end
		print(string.format("%-9s: sum=%4.1f -%s",
			(player.name or string.format("player%s", pid)),
			sum,
			tmp))
	end
end

function ai.evaluate_position_detailed()
	ai.evaluate_position(1)
	dump_evaluation_details()
end

local function calculate_portion_of_supply_given_by_building(g, bldgref)
	local rsrc_str = string.match(bldgref.actions[1], "^take (.*)$")
	assert(rsrc_str)

	local bag = resource_bag.parse(rsrc_str)
	return math.max(
		bag.food / g.food_supply,
		bag.wood / g.wood_supply,
		bag.stone / g.stone_supply,
		bag.cloth / g.cloth_supply)
end

local function estimate_value_of_building(g, position, bldgref)
	--estimate how many rounds are left
	local rounds_left = estimate_num_rounds_left()

	if bldgref.btype == "residence" then
		return 0
	end

	--estimate how many rounds before the building can be used
	local rounds_wait = math.max(0, position - state.bailiff)
	local factor = (rounds_left - rounds_wait) * .25

	if bldgref == WOOD_BUILDINGS[1]
		or bldgref == WOOD_BUILDINGS[2]
		or bldgref == WOOD_BUILDINGS[3]
		or bldgref == WOOD_BUILDINGS[4]
	then
		local portion_of_supply = calculate_portion_of_supply_given_by_building(g, bldgref)
		return factor * TUNABLE.VALUE_OF_WOODEN_PRODUCER * (portion_of_supply * 2)
	elseif bldgref == MASON then
		return factor * TUNABLE.VALUE_OF_MASON
	elseif bldgref == LAWYER then
		return factor * TUNABLE.VALUE_OF_LAWYER
	elseif bldgref == STONE_BUILDINGS[1]
		or bldgref == STONE_BUILDINGS[2]
		or bldgref == STONE_BUILDINGS[3]
	then
		local portion_of_supply = calculate_portion_of_supply_given_by_building(g, bldgref)
		return factor * TUNABLE.VALUE_OF_STONE_PRODUCER * (portion_of_supply * 2)
	elseif bldgref.btype == "wood" then
		return factor * TUNABLE.VALUE_OF_OTHER_WOODEN
	elseif bldgref.btype == "stone" then
		return factor * TUNABLE.VALUE_OF_OTHER_STONE
	else
		return 0
	end
end

local function value_of_favor(player_id)
	return 3
end

local function player_is_at_building(player_id, building)
	for i,pid in ipairs(building.occupants) do
		if pid == player_id then
			return true
		end
	end
	return false
end

local function evaluate_position_at_round_end(whoami, initial, params)
	if not initial then initial = {} end
	if not params then params = {} end

	local rounds_left = estimate_num_rounds_left()
	local est_score = {}

	--get input values that are the same for all players
	local g = {
		rounds_left = rounds_left,
		is_built = {},
		food_supply = 0,
		wood_supply = 0,
		stone_supply = 0,
		cloth_supply = 0,
		gold_supply = 0,
		money_supply = 0,
		}
	for i,bldg in ipairs(state.buildings) do
		if bldg.owner then
			g.is_built[bldg.ref] = i
		end
		if bldg.ref then
			local bags = {}
			if bldg.ref == PEDDLER1 then
				bags[1] = resource_bag.parse("3$")
			elseif bldg.ref == PEDDLER2 then
				bags[1] = resource_bag.parse("5$")
			elseif bldg.ref == MARKETPLACE1 then
				bags[1] = resource_bag.parse("1f")
				bags[2] = resource_bag.parse("1w")
				bags[3] = resource_bag.parse("1s")
				bags[4] = resource_bag.parse("1c")
			elseif bldg.ref.actions then
				for j,action_str in ipairs(bldg.ref.actions) do
					local rsrc_str = string.match(action_str, "^take (.*)$")
					if rsrc_str then
						table.insert(bags, resource_bag.parse(rsrc_str))
					end
				end
			end
			if #bags > 0 then
				local in_reach = 1/(1+math.exp(i-state.bailiff))
				local f = in_reach / #bags
				for j,bag in ipairs(bags) do
					g.food_supply = g.food_supply + bag.food * f
					g.wood_supply = g.wood_supply + bag.wood * f
					g.stone_supply = g.stone_supply + bag.stone * f
					g.cloth_supply = g.cloth_supply + bag.cloth * f
					g.gold_supply = g.gold_supply + bag.gold * f
					g.money_supply = g.money_supply + bag.money * f
				end
			end
		end
	end

--	if ai.VERBOSE then
--		print(string.format("estimate %.1f rounds left", rounds_left))
--	end
	if ai.DEBUG_GLOBAL_INPUT_VALUES then
		print(string.format(" supply of food is %.1f", g.food_supply))
		print(string.format(" supply of wood is %.1f", g.wood_supply))
		print(string.format(" supply of stone is %.1f", g.stone_supply))
		print(string.format(" supply of cloth is %.1f", g.cloth_supply))
		print(string.format(" supply of gold is %.1f", g.gold_supply))
		print(string.format(" supply of money is %.1f", g.money_supply))
	end

	EVALUATION = {}
	for player_id,me in pairs(state.players) do

		--
		--collect input values
		--

			local r = {
				food = me.resources.food,
				wood = me.resources.wood,
				stone = me.resources.stone,
				cloth = me.resources.cloth,
				gold = me.resources.gold,
				money = me.resources.money,
				points = me.score,
				}
			r.turn_order_position = position_in_turn_order(player_id)
			r.rounds_left = rounds_left
			r.contrib = estimate_maximum_contribution(r)
			r.income = me.income
			r.vp_favor = me.vp_favor
			r.money_favor = me.money_favor
			r.resource_favor = me.resource_favor
			r.build_favor = me.build_favor

			r.buildings = {}
			r.num_residences = 0
			for i = BRIDGE_POSITION+1, LAST_EMPTY_POSITION do
				local bldgref = state.buildings[i].ref
				if state.buildings[i].owner == player_id then
					if bldgref.btype == "residence" then
						r.num_residences = r.num_residences + 1
					else
						r.buildings[bldgref] = i
					end
				end
			end

		--
		--adjust input values if applicable
		--
			local my_castle_chances_left = r.rounds_left
			if params.in_truncated_round then
				my_castle_chances_left = my_castle_chances_left + 1
			end
			if my_castle_chances_left < 1 then
				r.contrib = 0
			elseif my_castle_chances_left < 2 then
				--having a bunch of resources set aside for
				--castle blocks doesn't help if we won't get
				--a chance to play them
				r.contrib = math.floor(r.contrib / r.turn_order_position)
			end

		--
		--assign values in terms of points at end of game
		--
		local eval = initial[player_id] or {}
		EVALUATION[player_id] = eval

		eval.vp = r.points

		--
		--estimate value of going early in the turn order
		--

			-- value of going early in the turn order
			if rounds_left > 0 and #state.order > 2 then
				eval.order = (#state.order - r.turn_order_position) * TUNABLE.VALUE_OF_EARLY_TURN_ORDER
			end

		--
		--estimate value of horded resources
		--

			-- value of unused gold
			eval.gold = 3 * r.gold

			-- value of potential castle blocks
			local worth_of_block = 3 + (r.rounds_left / 3)
			eval.contrib = worth_of_block * r.contrib
			
			-- value of other resources
			local n_resources = math.max(0,
				(r.food
				+ r.wood
				+ r.stone
				+ r.cloth) - 3 * r.contrib)
			local tmp_res_factor = TUNABLE.VALUE_OF_RESOURCES_PARAM1 / (n_resources + TUNABLE.VALUE_OF_RESOURCES_PARAM2)
			eval.resources = n_resources / 3
				+ n_resources * tmp_res_factor
				    * math.min(1,r.rounds_left / 4)

			-- TODO- value of having certain number of contributions
			-- in current castle phase

		--
		--estimate value of coins on hand
		--

			--  note: money not worth as much when you have a bunch
			--  note: if less than 2 rounds left, then money not
			--        worth as much either
			local tmp_money_factor = TUNABLE.VALUE_OF_MONEY_PARAM1 / (r.money + TUNABLE.VALUE_OF_MONEY_PARAM2)
			eval.money = r.money / 4
				+ r.money * tmp_money_factor
				    * math.min(1,r.rounds_left / 2)

		--	-- value of occupying the inn
		--	if rounds_left > 0 then
		--		eval.inn = 0
		--	end

		--
		--estimate value of buildings
		--

			-- value of owning buildings that other players use
			eval.buildings = 0
			for bldgref,bldg_position in pairs(r.buildings) do

				local bldg_value = estimate_value_of_building(g, bldg_position,bldgref)
				eval.buildings = eval.buildings + bldg_value
			end

		--
		--estimate potential value of not-yet-built prestige buildings
		--
			--prestige buildings require several things in place
			-- 1. gold (1-3)
			-- 2. stone (2-5)
			-- 3. if not stone, then cloth, wood, or food (3)
			-- 4. architect available
			-- 5. residences

			--check whether player has resources to build
			--a prestige building
			local residence_readiness =
				r.gold
				+ (r.stone / 2)
				+ math.max(r.cloth / 3, r.food / 3, r.wood / 3)
			local architect_on_hand
			if g.is_built[ARCHITECT1] or g.is_built[ARCHITECT2] then
				architect_on_hand = 1
			else
				architect_on_hand = 0
			end
			local lawyer_on_hand
			if g.is_built[LAWYER] then
				lawyer_on_hand = 1
			else
				lawyer_on_hand = 0
			end
			if rounds_left > 3 then
				--still time to build more residences
				r.num_residences = r.num_residences
					+ math.max(0, r.rounds_left / 4 + lawyer_on_hand - 1)
			end

			residence_readiness = residence_readiness
				+ (1 + architect_on_hand) * math.sqrt(r.num_residences)

			eval.residences = (TUNABLE.VALUE_OF_RESIDENCE * residence_readiness / 4)
				* math.min(1, r.rounds_left / 4)

		--
		--estimate value of income level
		--
			eval.income = r.income * r.rounds_left * 1.00

		--
		--estimate value of position on favor track
		--

			eval.favors = 0
			local FAVOR_POSITION_BONUS = { 4, 7, 9, 10, 10 }
			if r.vp_favor >= 1 then
				eval.favors = eval.favors + FAVOR_POSITION_BONUS[r.vp_favor]
			end
			if r.money_favor >= 1 then
				eval.favors = eval.favors + FAVOR_POSITION_BONUS[r.money_favor]
			end
			if r.resource_favor >= 1 then
				eval.favors = eval.favors + FAVOR_POSITION_BONUS[r.resource_favor]
			end
			if r.build_favor >= 1 then
				eval.favors = eval.favors + FAVOR_POSITION_BONUS[r.build_favor]
			end
			--value of position on favor track diminishes as game
			--progresses
			eval.favors = eval.favors * (r.rounds_left / 10)

		--end of all components

		local sum = 0
		for i,x in pairs(eval) do
			sum = sum + x
		end

		est_score[player_id] = sum

	end --loop through players

	return est_score
end

local function evaluate_scores(est_score, whoami)

	local sum = 0.0

	assert(#est_score == #state.order)
	assert(#est_score >= 2)

	for i,score in ipairs(est_score) do
		sum = sum + score
	end

	local avg_opponent_score = (sum - est_score[whoami])
			/ (#est_score - 1)
	return est_score[whoami] - avg_opponent_score

end

function ai.evaluate_position(whoami)

	local adjustment = {}
	local est_score = evaluate_position_at_round_end(whoami, adjustment)
	return evaluate_scores(est_score, whoami)

end

function ai.estimate_player_final_scores()
	return evaluate_position_at_round_end(1)
end

function quick_ai.select_builder_move()
	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

function quick_ai.select_provost_move()
	local whoami = current_turn()
	local is_free = (state.active_building == MERCHANT_GUILD_POSITION)
	if not is_free then
		return "noaction"
	end

	local last_my_worker = 0
	local last_opponent_worker = 0
	local need_to_save_coins = false
	for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do
		local bldg = state.buildings[i]
		if bldg.occupants[1] == whoami then
			last_my_worker = i
			if bldg.ref.id == "lawyer"
				or bldg.ref.id == "bank"
				or bldg.ref.id == "church"
				or bldg.ref.id == "wood-peddler"
				or bldg.ref == "fixed-peddler" then
				need_to_save_coins = true
			end
		elseif bldg.occupants[1] then
			last_opponent_worker = i
		end
	end
	if last_my_worker > state.bailiff - 3 then
		return "provost 3"
	else
		return "provost -3"
	end
end

local preferred_moves = {}
function quick_ai.select_move()
	local ph = current_phase()
	if ph == PHASE_WORKER_PLACEMENT then
		return "pass"
	elseif ph == PHASE_MOVE_PROVOST then
		return quick_ai.select_provost_move()
	else
		local whoami = current_turn()
		if state.active_building > 0 and state.active_building ~= BRIDGE_POSITION then
			local bldgref = state.buildings[state.active_building].ref
			if bldgref == LAWYER
				or bldgref == ARCHITECT1
				or bldgref == ARCHITECT2
				or bldgref == CARPENTER1
				or bldgref == CARPENTER2
			then
				return quick_ai.select_builder_move()
			end
			local key = string.format("%d,%d", whoami, state.active_building)
			local m = preferred_moves[key]
			if m and check_move(m) then
				return m
			end
			m = fallback_ai.select_move()
			preferred_moves[key] = m
			return m
		else
			return fallback_ai.select_move()
		end
	end
end

local function dump_choice(choice)
	local moves_string = choice.move
	local ee = choice.parent
	while ee.move do
		moves_string = ee.move .. "|" .. moves_string
		ee = ee.parent
	end
	if choice.principal_variation then
		moves_string = moves_string .. "   ("
		for i,m in ipairs(choice.principal_variation) do
			if i > 1 then moves_string = moves_string .. "|" end
			moves_string = moves_string .. m
		end
		moves_string = moves_string .. ")"
	end
	local MAX_LINE_LENGTH = 800
	if string.len(moves_string) > (MAX_LINE_LENGTH - 6) then
		moves_string = string.sub(moves_string, 1, (MAX_LINE_LENGTH - 8)) .. ".."
	end
	local my_score = evaluate_scores(choice.estimated_scores, context.whoami)
	print(string.format("%5.1f %s", my_score, moves_string))
end

local function new_node(parent)
	context.count_nodes = context.count_nodes + 1
	local choices = caylus.is_end_of_game(state) and {}
			or clone1(all_available_moves())
	return {
		state = state,
		alternate_moves = choices,
		evaluated = {},
		preferred_moves = {},
		estimated_scores = nil,
		parent = parent,
		turn = (#choices > 1 and current_turn() or 1),
		}
end

--check_alternatives(n)
--  n: a node of the game tree
--  returns: true if an alternative was found to be better
--     false otherwise
--
-- Creates a PV for each alternate choice at node n,
-- and updates node's n "preferred_moves" etc to reflect the
-- evaluated paths, and updates ancestors preferred moves
-- if necessary.
--
local function check_alternatives(n)

	if #n.alternate_moves == 0 then
		return false
	end

	local prior_best = n.preferred_moves[1]
	ai.evaluate_node(n, math.huge, 0)
	if prior_best == n.preferred_moves[1] then
		return false
	end

	if ai.VERBOSE then
		print(string.format("pv changed %s to %s",
				prior_best.move, n.preferred_moves[1].move))
	end
	return true
end

local function too_much_thinking()
	return context.count_nodes > 2000
	--return is_time_exceeded()
end

local function get_comparer(node)
	local function comparer(lhs,rhs)
		local lhs_score = evaluate_scores(lhs.estimated_scores, node.turn)
		local rhs_score = evaluate_scores(rhs.estimated_scores, node.turn)
		return lhs_score > rhs_score
	end
	return comparer
end

local function all_child_nodes_changed(node)

	local prior_best = assert(node.preferred_moves[1])
	local comparer = get_comparer(node)

	table.sort(node.preferred_moves, comparer)
	if prior_best ~= node.preferred_moves[1] then
		if ai.VERBOSE then
			print(string.format("pv changed %s to %s",
					prior_best.move, node.preferred_moves[1].move))
		end
	end

	node.estimated_scores = node.preferred_moves[1].estimated_scores
	return true
end

local function child_node_changed(node, idx)

	--TODO- since we know only one node's score changed,
	--perhaps a more optimal algorithm can be used.

	return all_child_nodes_changed(node)

end

local function refine_building_activation_plan(node)

	if not node.preferred_moves[1] then
		return false
	end

	if node.state.phase == PHASE_WORKER_PLACEMENT then
		if refine_building_activation_plan(node.preferred_moves[1]) then
			return child_node_changed(node, 1)
		end
		return false
	end

	local n = node
	if node.state.phase == PHASE_ACTIVATE_BUILDINGS
		and not too_much_thinking()
	then

		if ai.VERBOSE then
			print(string.format("examining what to do when %s (%d) is activated",
					n.state.buildings[n.state.active_building].ref.name,
					n.state.active_building))
		end
		if check_alternatives(node) then
			return true
		end
		n = n.preferred_moves[1]
	end
	return false
end

local function search_tree_for_preferred_move(node)
	if node.state.phase ~= PHASE_ACTIVATE_BUILDINGS then
		return nil
	end

	local n = context.root_node
	while n.preferred_moves[1]
			and n.state.phase == PHASE_WORKER_PLACEMENT do
		n = n.preferred_moves[1]
	end
	while n.preferred_moves[1]
			and n.state.phase == PHASE_ACTIVATE_BUILDINGS
			and n.state.active_building < node.state.active_building do
		n = n.preferred_moves[1]
	end

	while n.preferred_moves[1]
			and n.state.phase == PHASE_ACTIVATE_BUILDINGS
			and n.state.active_building == node.state.active_building
			and n.state.turn < node.state.turn do
		n = n.preferred_moves[1]
	end

	if n.preferred_moves[1]
			and n.state.phase == PHASE_ACTIVATE_BUILDINGS
			and n.state.active_building == node.state.active_building
			and n.state.turn == node.state.turn
	then
		return n.preferred_moves[1].move
	else
		return nil
	end
end

local function use_heuristics_to_order_alternates(node)

	local function analyze_this_move_next(movestr)
		for i,m in ipairs(node.alternate_moves) do
			if m == movestr then
				local tmp = node.alternate_moves[#node.alternate_moves]
				node.alternate_moves[#node.alternate_moves] = node.alternate_moves[i]
				node.alternate_moves[i] = tmp
				return true
			end
		end
		return false
	end

	local movestr
	--first, we use the search tree to find what move has been
	--favored at this level from another branch of the game tree

	movestr = search_tree_for_preferred_move(node)
	if movestr then
		if analyze_this_move_next(movestr) then
			return
		end
	end

	--that didn't work, so now we use our "quick_ai" to come up
	--with a preferred move
	movestr = quick_ai.select_move()
	analyze_this_move_next(movestr)
	return
end

local function node_has_psuedo_pass_ancestor(node)
	if node.move == "~pass" then
		return true
	elseif node.parent then
		return node_has_psuedo_pass_ancestor(node.parent)
	else
		return false
	end
end

local function skip_to_my_turn_again()
	assert(state.phase == PHASE_WORKER_PLACEMENT)
	assert(not state.players[context.whoami].passed)

	while state.order[state.turn] ~= context.whoami do
		next_turn()
	end
end

local function skip_to_building_activation_phase()

	local old_rule = NO_PASS_REWARD
	NO_PASS_REWARD = true
	while state.phase == PHASE_WORKER_PLACEMENT do
		do_move("pass")
	end
	NO_PASS_REWARD = old_rule
end

local function check_one_alternative_maybe(node,next_move)
	for i=1,#node.alternate_moves do
		if node.alternate_moves[i] == next_move then
			state = clone_game_state(node.state)
			table.remove(node.alternate_moves,i)
			do_move(next_move)
			local child_node = new_node(node)
			child_node.move = next_move
			ai.evaluate_node(child_node, 1, 0)
			node.evaluated[next_move] = child_node
			table.insert(node.preferred_moves, child_node)
			return child_node_changed(node, #node.preferred_moves)
		end
	end
	return false
end

local function report_choices(node)
	assert(ai.VERBOSE)
	
	state = node.state
	local _is_my_turn = is_my_turn(context.whoami)
	print(_is_my_turn and "my choices:" or "opponent choices:")
	for i,n in ipairs(node.preferred_moves) do
		dump_choice(n)
	end
end

local function generate_followup_moves(node,breadth,maxdepth)
	if maxdepth < 0 then
		debug("reached maxdepth")
		return false
	end
	if node.state.phase ~= PHASE_WORKER_PLACEMENT then
		return false
	end

	local child2 = node.preferred_moves[1]
	if not child2 then return false end

	debug("checking for alternatives to "..child2.move)

	if check_alternatives(child2) then
		local comparer = get_comparer(node)
		local favored_moves = {}
		favored_moves[child2.move] = 1
		favored_moves[child2.preferred_moves[1].move] = 1
		for i = 2, #node.preferred_moves do
			local child3 = node.preferred_moves[i]
			if comparer(child2, child3) then break end
			local child4 = child3.preferred_moves[1]
			for m in pairs(favored_moves) do
				check_one_alternative_maybe(child3, m)
			end
			if child3.preferred_moves[1] ~= child4 then
				--TODO
			end
			if i <= breadth then
				favored_moves[child3.move] = 1
			end
		end
		if ai.VERBOSE then
			report_choices(node)
		end
		return all_child_nodes_changed(node)
	end

	return false
end

function ai.evaluate_node(node, breadth, depth)

	assert(context.whoami)
	assert(context.is_stable_state)

	state = node.state

	if caylus.is_end_of_game(state) or context.is_stable_state() then
		--new round; we can evaluate here
		if not node.estimated_scores then
			node.estimated_scores = evaluate_position_at_round_end(context.whoami,
					{},
					{ in_truncated_round = node_has_psuedo_pass_ancestor(node), }
					)

			if ai.VERBOSE then
				dump_choice(node)
				--dump_evaluation_details()
			end
		end
		return node.estimated_scores, nil
	end

	--we maximize for the player whose turn it is

	if #node.preferred_moves == 0 then

		--no moves have been evaluated yet
		--apply some heuristics to determine which move
		--we will look at first

		if #node.alternate_moves > 1 then
			use_heuristics_to_order_alternates(node)
		else
			assert(#node.alternate_moves >= 1)
		end
	end

	if #node.alternate_moves == 0 then

		--all moves have been evaluated
		return node.estimated_score, node.preferred_moves[1].move
	end

	while #node.preferred_moves < breadth
		and #node.alternate_moves > 0
		and not (#node.preferred_moves >= 1 and too_much_thinking())
	do

		local next_move

		state = clone_game_state(node.state)

			next_move = table.remove(node.alternate_moves)
			do_move(next_move)

		local child_node = new_node(node)
		child_node.move = next_move

		ai.evaluate_node(child_node, 1, 0)
		node.evaluated[next_move] = child_node

		--FIXME- better to just insert at the correct spot
		table.insert(node.preferred_moves, child_node)
		child_node_changed(node, #node.preferred_moves)

	end --loop through available choices at this node

	if ai.VERBOSE and #node.preferred_moves > 1 then
		report_choices(node)
	end

	state = node.state
	if depth >= 1 then

		--check that the "principal variation" we're using to
		--determine the valuation of the move is really the
		--best sequence of moves

		--if we find the principal variation changes, we
		--repeat the process with the adjusted variation until
		--we settle on an accurate variation

		local function refine_first_alternates_pv()
			if #node.preferred_moves >= 2
				and refine_building_activation_plan(node.preferred_moves[2])
			then
				return child_node_changed(node, 2)
			end
			return false
		end

		while not too_much_thinking()
		do

			if not (
				refine_building_activation_plan(node) or
				--refine_first_alternates_pv() or
				generate_followup_moves(node,3,1)
				) then

				break
			end

		end --loop
	end

--	if breadth > 1 and depth >= 1 then
--		local max_to_look_at = 3
--		local new_results = {}
--		for i = 1, math.min(3, #node.preferred_moves) do
--
--			local n = node.preferred_moves[i]
--			if ai.VERBOSE then
--				print("examining follow-ups for "..n.move)
--			end
--			ai.evaluate_node(n, math.huge, depth - 1)
--			table.insert(new_results, n)
--		end
--
--		node.preferred_moves = new_results
--		table.sort(node.preferred_moves, COMPARER)
--
--		if ai.VERBOSE then
--			print(string.format("%s, after looking one move ahead:",
--				(_is_my_turn and "my choices" or "opponent choices")))
--			for i,n in ipairs(node.preferred_moves) do
--				dump_choice(n)
--			end
--		end
--	end

	local best = node.preferred_moves[1]
	node.estimated_scores = best.estimated_scores
	return node.estimated_scores, best.move

end

function ai.select_move()

	start_thinking()
	context = {
		whoami = current_turn(),
		count_nodes = 0,
		}
	local saved_state = state

	if state.phase == PHASE_SCORING then
		context.is_stable_state = function()
				return state.round > saved_state.round
			end
	else
		context.is_stable_state = function()
				return state.phase == PHASE_SCORING
					or state.round > saved_state.round
			end
	end

	context.root_node = new_node(nil)

	local next_move

	if #context.root_node.alternate_moves == 1 then
		next_move = context.root_node.alternate_moves[1]
	else
		ai.evaluate_node(context.root_node, math.huge, 1);
		next_move = context.root_node.preferred_moves[1].move
	end

	--finish_thinking()
	--print(string.format("%.2fs / %d nodes / %s", get_elapsed(), context.count_nodes, next_move))

	state = saved_state
	if not ai.SAVE_ANALYSIS then
		context = nil
		collectgarbage("collect")
	end

	return next_move

end

function ai.show_move_evaluation(movestr)
	local saved_state = state
	local node = context.root_node.evaluated[movestr]
	assert(node)

	--follow principal variation to end
	while node.preferred_moves[1] do
		node = node.preferred_moves[1]
	end

	dump_choice(node)
	state = node.state
	ai.evaluate_position_detailed()

	state = saved_state
	return
end

return ai
