--BEGIN NEURAL NET LIBRARY
local NNlib = {}
do

function NNlib.prepare(WEIGHTS)

	local FORWARD_LINKS = {}
	for nodename,node in pairs(WEIGHTS) do
		for srcnode in pairs(node) do
			FORWARD_LINKS[srcnode] = (FORWARD_LINKS[srcnode] or {})
			FORWARD_LINKS[srcnode][nodename] = true
		end
	end
	local NN = {
		weights = WEIGHTS,
		forward_links = FORWARD_LINKS,
		}
	return NN
end

local function sum_inputs_default(NN, eval_struct, node_name)
	local rv = 0
	for src_node_name, weight in pairs(NN.weights[node_name]) do
		assert(eval_struct[src_node_name], "node "..src_node_name.." is nil")
		rv = rv + eval_struct[src_node_name] * weight
	end
	return rv
end
function NNlib.sum_inputs(NN, node_values, node_name)
	local impl = NN.functions[node_name].sum_inputs or sum_inputs_default
	return impl(NN, node_values, node_name)
end

function NNlib.evaluate(NN, input_values)

	local eval = {}
	for k,v in pairs(input_values) do
		eval[k] = v
	end

	--here we do foreward propagation
	local function calculate_value_of(node_name)
		if eval[node_name] then return eval[node_name] end
		if not NN.weights[node_name] then return 0 end

		for srcnode in pairs(NN.weights[node_name]) do
			eval[srcnode] = calculate_value_of(srcnode)
		end
		local e = NNlib.sum_inputs(NN, eval, node_name)
		return NN.functions[node_name].activator(e)
	end
	eval.score = calculate_value_of("score")

	return eval
end

function NNlib.do_backpropagation(NN, outputs, output_node, output_error)

	--error_signals is a hashtable, indexed by node name
	-- the values are the desired increase in value going into
	-- that node

	local error_signals = {}
	error_signals[output_node] = output_error

	local calculate_error_signal_on_edge
	local function calculate_error_signal_for(node)
		if not error_signals[node] then
			local s = 0
			for destnode in pairs(NN.forward_links[node]) do
				s = s + calculate_error_signal_on_edge(destnode,node)
			end

			if NN.functions[node] then

				--use the derivative of node's activator to
				--adjust the error signal here
				local e = NNlib.sum_inputs(NN, outputs, node)
				local d = NN.functions[node].derivative(e)

				s = s * d
			end

			error_signals[node] = s
		end
		return error_signals[node]
	end
	function calculate_error_signal_on_edge(destnode,srcnode)
		if NN.functions[destnode].is_product then
			local z = calculate_error_signal_for(destnode)
			local e = outputs[srcnode]
			local d = 1/math.abs(e)
			return z * d
		end
		return NN.weights[destnode][srcnode] * calculate_error_signal_for(destnode)
	end

	for node in pairs(NN.forward_links) do
		calculate_error_signal_for(node)
	end

--	--dump error signals
--	for node_name, amt in pairs(error_signals) do
--		print(string.format("%-20s: %.2f", node_name, amt))
--	end

	--calculate weight adjustments
	local weight_adjustments = {}
	for node_name in pairs(NN.weights) do
		weight_adjustments[node_name] = {}
		for srcnode, weight in pairs(NN.weights[node_name]) do
			weight_adjustments[node_name][srcnode] = error_signals[node_name] * outputs[srcnode]
		end
	end
	return {
		outputs = outputs,
		error_signals = error_signals,
		adjustments = weight_adjustments,
		}
end

	local function _product_inputs(NN, node_values, node_name)
		local rv = 0
		for src_node_name in pairs(NN.weights[node_name]) do
			local t = node_values[src_node_name]
			rv = rv + math.log(math.abs(t))
		end
		return rv
	end
	local function _tanh(x)
		local Z = math.exp(2*x)
		return (Z-1)/(Z+1)
	end
NNlib.nodetypes = {
	linearfn = {
		activator = function(x) return x end,
		derivative = function(x) return 1 end,
		inverse = function(x) return x end,
		},
	productfn = {
		activator = function(x) return math.exp(x) end,
		derivative = function(x) return math.exp(x) end,
		sum_inputs = _product_inputs,
		is_product = true,
		},
	exp_fn = {
		activator = function(x) return math.exp(x) end,
		derivative = function(x) return math.exp(x) end,
		inverse = function(x) return math.log(x) end,
		},
	log_fn = {
		activator = function(x) return math.log(x) end,
		derivative = function(x) return 1/x end,
		inverse = function(x) return math.exp(x) end,
		},
	tanh_fn = {
		activator = _tanh,
		derivative = function(x) return 1-math.pow(_tanh(x),2) end,
		},
	}

end --lexical block
--END NEURAL NET LIBRARY

local OPPONENT_SCORE_WEIGHT = 1
local WEIGHTS = {}
WEIGHTS.score =
	{
		vp = 1.0,
		order = 0.7,
		gold = 3.0,
		resources_value = 0.42,
		money_value = 1.0,
		inn = 1.0,
		buildings = 0.82,
		residences = 3.60,
		income_value = 1.10,
		favors_value = .01,
		blocks_in_phase = 2.0,
	}
WEIGHTS.income_value = {
		income = 1.0,
		rounds_left = 1.0,
	}
WEIGHTS.favors_value = {
		favors = 1.0,
		rounds_left = 1.0,
	}
WEIGHTS.favors = {
		vp_favor = 1.0,
		money_favor = 1.0,
		resource_favor = 1.0,
		build_favor = 1.0,
	}
WEIGHTS.money_value = {
		money_T = 6.805,
	}
WEIGHTS.money_T = {
		money = .21,
		bias = 0.04,
	}
WEIGHTS.resources_value0 = {
		n_resources = -0.85,
		has_3_cubes = 3,
		has_6_cubes = 1.5,
		has_food_and_wood_T = 3,
		has_food_and_stone_T = 3,
		n_food_T = 3,
		n_wood_T = 3,
		n_stone_T = 3,
		n_cloth_T = 3,
		contrib = 0.71,
	}
WEIGHTS.resources_value1 = {
		resources_value0 = 1,
		rounds_left = 1,
	}
WEIGHTS.resources_value = {
		--resources_value0 = -.41,
		--resources_value1 = .27,
		--bias = -.051,
		resources_value0 = 1,
		resources_value1 = 0,
		bias = 0,
	}
WEIGHTS.has_food_and_wood_T = { has_food_and_wood = .333 }
WEIGHTS.has_food_and_stone_T = { has_food_and_stone = .333 }
WEIGHTS.n_food_T = { n_food = .333 }
WEIGHTS.n_wood_T = { n_wood = .333 }
WEIGHTS.n_stone_T = { n_stone = .333 }
WEIGHTS.n_cloth_T = { n_cloth = .333 }
WEIGHTS.buildings = {
		--wood buildings that players can build
		["has_building_wood-2w"] = 0.80,
		["has_building_wood-2f"] = 0.80,
		["has_building_wood-2c"] = 1.1,
		["has_building_wood-2s"] = 0.9,
		["has_building_wood-peddler"] = 0.10,
		["has_building_wood-market"] = 0.11,
		has_building_mason = 0.32,
		has_building_lawyer = 0.41,
		--stone buildings
		has_building_architect = 0.25,
		has_building_bank = 0.25,
		has_building_alchemist = 0.25,
		has_building_church = 0.25,
		has_building_tailor = 0.25,
		has_building_jeweler = 0.25,
		["has_building_stone-ffc"] = 0.9,
		["has_building_stone-wwf"] = 0.9,
		["has_building_stone-ssc"] = 1.1,
	}

local NN = NNlib.prepare(WEIGHTS)
NN.functions = {
	score                = NNlib.nodetypes.linearfn,
	income_value         = NNlib.nodetypes.productfn,
	favors_value         = NNlib.nodetypes.productfn,
	favors               = NNlib.nodetypes.linearfn,
	money_value          = NNlib.nodetypes.linearfn,
	resources_value0     = NNlib.nodetypes.linearfn,
	resources_value1     = NNlib.nodetypes.productfn,
	resources_value      = NNlib.nodetypes.linearfn,
	buildings            = NNlib.nodetypes.linearfn,
	has_food_and_wood_T  = NNlib.nodetypes.tanh_fn,
	has_food_and_stone_T = NNlib.nodetypes.tanh_fn,
	n_food_T             = NNlib.nodetypes.tanh_fn,
	n_wood_T             = NNlib.nodetypes.tanh_fn,
	n_stone_T            = NNlib.nodetypes.tanh_fn,
	n_cloth_T            = NNlib.nodetypes.tanh_fn,
	money_T              = NNlib.nodetypes.tanh_fn,
	}

local valuation = {
	WEIGHTS = WEIGHTS,
	}
local function estimate_maximum_contribution(r)

	--FIXME: with 3f,2w,2s,2c then it is possible to make 3 contributions,
	--but this function does not recognize this

	return math.max(
			math.min(r.food, r.wood, r.stone + r.cloth),
			math.min(r.food, r.stone, r.wood + r.cloth),
			math.min(r.food, r.cloth, r.wood + r.stone)
			)
end

function valuation.estimate_num_rounds_left()
	if state.castle_phase > 3 then
		return 0
	end
	local TYPICAL_NUMBER_OF_ROUNDS = 14.7

	local bailiff_left = CASTLE_BAILIFF_THRESHOLDS[3] - state.bailiff 
	local castle_left = CASTLE_CONTRIBUTION_THRESHOLDS[3] - state.contribution_count

	local bailiff_portion = (state.bailiff - BAILIFF_START_POSITION)
		/ (CASTLE_BAILIFF_THRESHOLDS[3] - BAILIFF_START_POSITION)
	local castle_portion = state.contribution_count / CASTLE_CONTRIBUTION_THRESHOLDS[3]

	local est_castle_rounds = castle_portion > 0 and state.round / castle_portion or 15
	local est_bailiff_rounds = bailiff_portion > 0 and state.round / bailiff_portion or 15
	local est_rounds = math.min(est_castle_rounds, est_bailiff_rounds)
	local est_rounds_left = est_rounds - state.round

	if bailiff_left <= 1 or castle_left <= 4 then
		return math.min(1, est_rounds_left)
	end
	if bailiff_left <= 3 or castle_left <= 7 then
		return math.min(2, est_rounds_left)
	end
	if state.castle_phase == 3 then
		return math.min(5, est_rounds_left)
	elseif state.castle_phase == 2 then
		local est1 = TYPICAL_NUMBER_OF_ROUNDS - (state.round - 1)
		local est2 = math.min(9, est_rounds_left)
		return (est1 + est2) / 2
	elseif state.castle_phase == 1 then
		return TYPICAL_NUMBER_OF_ROUNDS - (state.round - 1)
	end

	return est_rounds_left
end

local EVALUATION
local function dump_evaluation_details()
	--for pid,player in pairs(state.players) do
	for pid = 1,1 do
		local player = state.players[pid]
		local eval = EVALUATION[pid]
		local tmp = ""
		for n,w in pairs(WEIGHTS.score) do
			tmp = tmp .. string.format(" %s=%.1f", n, w * eval[n])
		end
		print(string.format("%-9s: sum=%4.1f -%s",
			(player.name or string.format("player%s", pid)),
			eval.score,
			tmp))
	--drill down components of resources_value
	--	for n,w in pairs(WEIGHTS.resources_value) do
	--		print(string.format("        %-38s:%10.4f",
	--			n .. "/resources_value",
	--			w * eval[n]))
	--	end
	end
end

function valuation.evaluate_position_detailed()
	valuation.evaluate_position(1)
	dump_evaluation_details()
end

local function calculate_portion_of_supply_given_by_building(g, bldgref)
	local rsrc_str = string.match(bldgref.actions[1], "^take (.*)$")
	assert(rsrc_str)

	local bag = resource_bag.parse(rsrc_str)
	return math.max(
		bag.food / g.food_supply,
		bag.wood / g.wood_supply,
		bag.stone / g.stone_supply,
		bag.cloth / g.cloth_supply)
end

local function value_of_favor(player_id)
	return 3
end

function valuation.evaluate_position_at_round_end(whoami, initial, params)
	if not initial then initial = {} end
	if not params then params = {} end

	local rounds_left = valuation.estimate_num_rounds_left()
	local est_score = {}

	--get input values that are the same for all players
	local g = {
		rounds_left = rounds_left,
		is_built = {},
		food_supply = 0,
		wood_supply = 0,
		stone_supply = 0,
		cloth_supply = 0,
		gold_supply = 0,
		money_supply = 0,
		}
	for i,bldg in ipairs(state.buildings) do
		if bldg.owner then
			g.is_built[bldg.ref] = i
		end
		if bldg.ref then
			local bags = {}
			if bldg.ref == PEDDLER1 then
				bags[1] = resource_bag.parse("3$")
			elseif bldg.ref == PEDDLER2 then
				bags[1] = resource_bag.parse("5$")
			elseif bldg.ref == MARKETPLACE1 then
				bags[1] = resource_bag.parse("1f")
				bags[2] = resource_bag.parse("1w")
				bags[3] = resource_bag.parse("1s")
				bags[4] = resource_bag.parse("1c")
			elseif bldg.ref.actions then
				for j,action_str in ipairs(bldg.ref.actions) do
					local rsrc_str = string.match(action_str, "^take (.*)$")
					if rsrc_str then
						table.insert(bags, resource_bag.parse(rsrc_str))
					end
				end
			end
			if #bags > 0 then
				local in_reach = 1/(1+math.exp(i-state.bailiff))
				local f = in_reach / #bags
				for j,bag in ipairs(bags) do
					g.food_supply = g.food_supply + bag.food * f
					g.wood_supply = g.wood_supply + bag.wood * f
					g.stone_supply = g.stone_supply + bag.stone * f
					g.cloth_supply = g.cloth_supply + bag.cloth * f
					g.gold_supply = g.gold_supply + bag.gold * f
					g.money_supply = g.money_supply + bag.money * f
				end
			end
		end
	end

--	if ai.VERBOSE then
--		print(string.format("estimate %.1f rounds left", rounds_left))
--	end
	if ai.DEBUG_GLOBAL_INPUT_VALUES then
		print(string.format(" supply of food is %.1f", g.food_supply))
		print(string.format(" supply of wood is %.1f", g.wood_supply))
		print(string.format(" supply of stone is %.1f", g.stone_supply))
		print(string.format(" supply of cloth is %.1f", g.cloth_supply))
		print(string.format(" supply of gold is %.1f", g.gold_supply))
		print(string.format(" supply of money is %.1f", g.money_supply))
	end

	EVALUATION = {}
	for player_id,me in pairs(state.players) do

		--
		--collect input values
		--

			local r = {
				food = me.resources.food,
				wood = me.resources.wood,
				stone = me.resources.stone,
				cloth = me.resources.cloth,
				gold = me.resources.gold,
				money = me.resources.money,
				points = me.score,
				}
			r.turn_order_position = position_in_turn_order(player_id)
			r.rounds_left = rounds_left
			r.contrib = estimate_maximum_contribution(r)
			r.income = me.income
			r.vp_favor = me.vp_favor
			r.money_favor = me.money_favor
			r.resource_favor = me.resource_favor
			r.build_favor = me.build_favor

			r.buildings = {}
			r.num_residences = 0
			for i = BRIDGE_POSITION+1, LAST_EMPTY_POSITION do
				local bldgref = state.buildings[i].ref
				if state.buildings[i].owner == player_id then
					if bldgref.btype == "residence" then
						r.num_residences = r.num_residences + 1
					else
						r.buildings[bldgref] = i
					end
				end
			end

		--
		--count how many blocks this player has contributed to the
		--castle during this phase
		--
			r.blocks_in_phase = 0
			local i0 = (CASTLE_CONTRIBUTION_THRESHOLDS[state.castle_phase - 1] or 0) + 1
			local i1 = (CASTLE_CONTRIBUTION_THRESHOLDS[state.castle_phase]) or 0
			for i = i0, i1 do
				if state.contributions[i] == player_id then
					r.blocks_in_phase = r.blocks_in_phase + 1
				end
			end

		--
		--adjust input values if applicable
		--
			local my_castle_chances_left = r.rounds_left
			if params.in_truncated_round then
				my_castle_chances_left = my_castle_chances_left + 1
			end
			if my_castle_chances_left < 1 then
				r.contrib = 0
			elseif my_castle_chances_left < 2 then
				--having a bunch of resources set aside for
				--castle blocks doesn't help if we won't get
				--a chance to play them
				r.contrib = math.floor(r.contrib / r.turn_order_position)
			end

		--
		--assign values in terms of points at end of game
		--
		local eval = initial[player_id] or {}

		eval.vp = r.points

		--
		--estimate value of going early in the turn order
		--

			-- value of going early in the turn order
			if rounds_left > 0 and #state.order > 2 then
				eval.order = #state.order - r.turn_order_position
			else
				eval.order = 0
			end

		--
		--estimate value of horded resources
		--

			-- value of unused gold
			eval.gold = r.gold

			-- value of potential castle blocks
			local worth_of_block = 3 + (r.rounds_left / 3)
			eval.contrib = worth_of_block * r.contrib
			
			-- value of other resources
			local n_resources = math.max(0,
				(r.food
				+ r.wood
				+ r.stone
				+ r.cloth) - 3 * r.contrib)
			eval.n_food = r.food
			eval.n_wood = r.wood
			eval.n_stone = r.stone
			eval.n_cloth = r.cloth
			eval.n_resources = r.food + r.wood + r.stone + r.cloth
			eval.has_3_cubes = 1/(1+math.exp(-(eval.n_resources-3)))
			eval.has_6_cubes = 1/(1+math.exp(-(eval.n_resources-6)))

			-- value of blocks in castle but not yet scored
			-- (i.e. end of phase favors)
			eval.blocks_in_phase = r.blocks_in_phase

			eval.has_food_and_wood = math.min(r.food,r.wood)
			eval.has_food_and_stone = math.min(r.food,r.stone)

		--
		--estimate value of coins on hand
		--

			--  note: money not worth as much when you have a bunch
			--  note: if less than 2 rounds left, then money not
			--        worth as much either
			--FIXME: no consideration of diminishing value of
			--money at end of game

			eval.money = r.money

			-- value of occupying the inn
			eval.inn = 0
		--	if rounds_left > 0 then
		--		eval.inn = 0
		--	end

		--
		--estimate value of buildings
		--

			-- value of owning buildings that other players use
			for bldgref,bldg_position in pairs(r.buildings) do

				local rounds_wait = math.max(0,bldg_position-state.bailiff)
				local rounds_in_play = g.rounds_left - rounds_wait

				eval["has_building_"..bldgref.id] = rounds_in_play
			end

		--
		--estimate potential value of not-yet-built prestige buildings
		--
			--prestige buildings require several things in place
			-- 1. gold (1-3)
			-- 2. stone (2-5)
			-- 3. if not stone, then cloth, wood, or food (3)
			-- 4. architect available
			-- 5. residences

			eval.residences = 0
			for i = 1, r.num_residences do
				if g.rounds_left >= 3 * (i-1) then
					eval.residences = eval.residences + 1 / i
				end
			end

		--
		--estimate value of income level
		--
			eval.income = r.income
			eval.rounds_left = r.rounds_left

		--
		--estimate value of position on favor track
		--

			local FAVOR_POSITION_BONUS = { 4, 7, 9, 10, 10 }
			if r.vp_favor >= 1 then
				eval.vp_favor = FAVOR_POSITION_BONUS[r.vp_favor]
			end
			if r.money_favor >= 1 then
				eval.money_favor = FAVOR_POSITION_BONUS[r.money_favor]
			end
			if r.resource_favor >= 1 then
				eval.resource_favor = FAVOR_POSITION_BONUS[r.resource_favor]
			end
			if r.build_favor >= 1 then
				eval.build_favor = FAVOR_POSITION_BONUS[r.build_favor]
			end

		--bias unit
			eval.bias = 1

		--end of all components


		--here we do foreward propagation
		eval = NNlib.evaluate(NN, eval)
		est_score[player_id] = eval.score
		EVALUATION[player_id] = eval

	end --loop through players

	return est_score
end

function valuation.evaluate_scores(est_score, whoami)

	local sum = 0.0

	assert(#est_score == #state.order)
	assert(#est_score >= 2)

	for i,score in ipairs(est_score) do
		sum = sum + score
	end

	local avg_opponent_score = (sum - est_score[whoami])
			/ (#est_score - 1)
	return est_score[whoami] - OPPONENT_SCORE_WEIGHT * avg_opponent_score

end

function valuation.evaluate_position(whoami)

	local adjustment = {}
	local est_score = valuation.evaluate_position_at_round_end(whoami, adjustment)
	return valuation.evaluate_scores(est_score, whoami)

end

function valuation.estimate_player_final_scores()
	return valuation.evaluate_position_at_round_end(1)
end

--@param final_scores an array of actual game-end scores, indexed by player id
--
function valuation.help_learn(real_scores)
	valuation.evaluate_position_at_round_end(1)
	local rv = {}
	for pid in pairs(real_scores) do
		local eval = EVALUATION[pid]
		rv[pid] = NNlib.do_backpropagation(NN, eval, "score", real_scores[pid] - eval.score)
	end
	return rv
end

return valuation
